/**
 * gclc-socket:fr.bigeon.gclc.socket.PlugableConsoleInput.java
 * Created on: Nov 18, 2017
 */
package fr.bigeon.gclc.socket;

/*-
 * #%L
 * GCLC Socket
 * %%
 * Copyright (C) 2016 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.manager.ConsoleInput;
import fr.bigeon.gclc.manager.ReadingRunnable;
import fr.bigeon.gclc.tools.ConstantString;
import fr.bigeon.gclc.tools.StringProvider;

/** A console input where the stream can be plugged.
 * <p>
 * This pluggable console input accepts an input and output to be connected to
 * it. The connexion cannot be concurrent, which mean that any connected stream
 * must be disconnected before a new call to
 * {@link #connect(InputStream, PrintStream)} is done.
 *
 * @author Emmanuel Bigeon */
public final class PluggableConsoleInput implements ConsoleInput {
    /** The ten constant. */
    private static final int TENTH = 10;
    /** Class logger. */
    private static final Logger LOGGER = Logger
            .getLogger(PluggableConsoleInput.class.getName());
    /** The default time out. */
    private static final long TIMEOUT = 100;
    /** The prompting. */
    private boolean prompting = false;
    /** If the element is closed. */
    private boolean closed = false;
    /** The default prompt. */
    private StringProvider prompt = new ConstantString("> "); //$NON-NLS-1$
    /** If the input is plugged or buffering. */
    private boolean connected = false;
    /** The current connexion (if any). */
    private ReadingRunnable connexion;
    /** The interrupted status for prompts. */
    private boolean interrupted = false;
    /** The last hint hint. */
    private String hint;
    /** The output for hints. */
    private PrintStream output;

    // Locks
    /** The lock for connexion and disconnexion of actual streams. */
    private final Object connexionLock = new Object();

    /** Create the pluggable console input. */
    public PluggableConsoleInput() {
        // do nothing
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#close() */
    @Override
    public void close() {
        closed = true;
    }

    /** Connect the given input stream to the input and output to the hints
     * writing.
     *
     * @param stream the input stream
     * @param out the output for hints.
     * @throws IOException if the input is already connected. */
    public void connect(final InputStream stream,
                        final PrintStream out) throws IOException {
        synchronized (connexionLock) {
            if (connected) {
                throw new IOException(
                        "Input already connected to an input stream"); //$NON-NLS-1$
            }

            output = out;
            if (prompting) {
                out.print(hint);
                out.flush();
            }

            final InputStreamReader streamReader = new InputStreamReader(
                    stream, StandardCharsets.UTF_8);
            final BufferedReader reader = new BufferedReader(streamReader);
            connexion = new ReadingRunnable(reader);
            final Thread th = new Thread(connexion, "GCLC Socket - Read input"); //$NON-NLS-1$
            th.start();
            connexionLock.notifyAll();
            connected = true;
        }
    }

    /** Disconnect the current input and hint output. */
    public synchronized void disconnect() {
        synchronized (connexionLock) {
            if (!connected) {
                return;
            }
            connected = false;
            connexion.setRunning(false);
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#getPrompt() */
    @Override
    public StringProvider getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        interrupted = true;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#isClosed() */
    @Override
    public boolean isClosed() {
        return closed;
    }

    /** Test if a prompt is occuring.
     *
     * @return the prompting */
    public boolean isPrompting() {
        return prompting;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt() */
    @Override
    public String prompt() throws IOException {
        return prompt(prompt.apply());
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(long) */
    @Override
    public String prompt(final long timeout) throws IOException {
        return prompt(prompt.apply(), timeout);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) throws IOException {
        if (closed) {
            throw new IOException();
        }
        prompting = true;
        hint = message;
        synchronized (connexionLock) {
            hint = message;
            if (connected) {
                output.print(message);
                output.flush();
            }
        }

        String res = null;
        while (res == null && !interrupted) {
            try {
                res = waitMessageOrConnexion(TIMEOUT, TIMEOUT / TENTH);
            } catch (final InterruptedException e) {
                LOGGER.log(Level.FINE, "Interruption of thread", e); //$NON-NLS-1$
                Thread.currentThread().interrupt();
            }
            if (closed) {
                throw new IOException();
            }
        }
        prompting = false;
        return res;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String,
     * long) */
    @Override
    public String prompt(final String message,
                         final long timeout) throws IOException {
        if (closed) {
            throw new IOException();
        }
        prompting = true;
        synchronized (connexionLock) {
            hint = message;
            if (connected) {
                output.print(message);
                output.flush();
            }
        }

        String res = null;
        final long tic = System.currentTimeMillis();
        long time = System.currentTimeMillis() - tic;
        while (res == null && !interrupted && time < timeout) {
            try {
                res = waitMessageOrConnexion(timeout - time,
                        (timeout - time) / TENTH);
            } catch (final InterruptedException e) {
                LOGGER.log(Level.FINE, "Interruption of thread", e); //$NON-NLS-1$
                Thread.currentThread().interrupt();
            }
            time = System.currentTimeMillis() - tic;
            if (closed) {
                throw new IOException();
            }
        }
        prompting = false;
        return res;
    }

    @Override
    public void setPrompt(final StringProvider prompt) {
        this.prompt = prompt;
    }
    
    @Override
    public void setPrompt(String prompt) {
    	setPrompt(new ConstantString(prompt));
    }

    /** Wait for a hint or connexion.
     *
     * @param messageTimeout the timeout on the current connexion hint waiting
     * @param connexionTimeout the timeout on the new connexion wait
     * @return the hint, or null if not connected or timed out.
     * @throws IOException if the reading failed.
     * @throws InterruptedException if the wait was interrupted */
    private String waitMessageOrConnexion(final long messageTimeout,
                                          final long connexionTimeout) throws IOException,
                                                                       InterruptedException {
        synchronized (connexionLock) {
            if (connected) {
                return connexion.getNextMessage(messageTimeout);
            }
            connexionLock.wait(connexionTimeout);
        }
        return null;
    }
}
